/******************************************************************************\
*
* File:          VoxelDemoA.cpp
* Creation date: September 07, 2003 23:02
* Author:        ClassBuilder
*                XXXX
* Purpose:       Method implementations of class 'VoxelDemo'
*
* Modifications: @INSERT_MODIFICATIONS(* )
* September 07, 2003 23:02 Initial
*     Updated code of method 'InitTables'
*     Updated code of method 'InitHeightMap'
*     Updated code of method 'GameLoop'
*     Updated code of method 'GameInit'
*     Updated code of method 'GameEnd'
*     Updated code of method 'DrawTerrain'
*     Updated code of method '~VoxelDemo'
*     Updated code of method 'VoxelDemo'
*     Updated member 'Depth'
*     Updated member 'Width'
*     Updated member 'HeightBuffer'
*     Updated member 'TextureBuffer'
*     Updated member 'BackBuffer'
*     Updated member 'HeightData'
*     Updated member 'TextureData'
*     Updated member 'm_display'
*
* Copyright 2003, XXXXX
* All rights are reserved. Reproduction in whole or part is prohibited
* without the written consent of the copyright owner.
*
\******************************************************************************/
//@START_USER1
//@END_USER1


// Master include file
#include "VoxelMaster.h"


//@START_USER2
CComModule _Module;

float sin_look[ANGLE_360];
float cos_look[ANGLE_360];
float sphere_cancel[ANGLE_60];

float nearStepSize;
float farStepSize;

float viewAngle = 60.0f * 3.14159265f / 180.0f;

int numberOfSlices = 100;
int nearScreenRow = 200;
int farScreenRow = nearScreenRow - numberOfSlices;

int screen_z = 0;

int play_x = 1000;
int play_y = 1000;
int play_z = 275;
int play_ang = ANGLE_0;
int play_dist = 70;
int mountain_scale = 60;

int last_top[DISP_X];

unsigned char heightMap[512*512];

HANDLE hFile;
DWORD temp;

//@END_USER2


using namespace Frog;





//
// Group: Constructors and Destructor
//
/*@NOTE_1181
Constructor method.
*/
VoxelDemo::VoxelDemo() //@INIT_1181
{//@CODE_1181

    m_config.szWindowName     = _T("Texture Sample");
    m_config.orientation      = ORIENTATION_NORTH;
    m_config.splashScreenTime = 0;

}//@CODE_1181


/*@NOTE_1177
Destructor method.
*/
VoxelDemo::~VoxelDemo()
{//@CODE_1177
    DestructorInclude();

    // Put in your own code
}//@CODE_1177


// End of Group: Constructors and Destructor


//
// Group: Initialization Methods
//
void VoxelDemo::InitHeightMap()
{//@CODE_1192

	int min = 0xFFFFFF, max = 0;
	Pixel *buffer = HeightBuffer.pixels;
	for (int i=0; i<512; i++) {
		for (int j=0; j<512; j++) {
			if (*buffer > max) max = *buffer;
			if (*buffer < min) min = *buffer;
			buffer++;
		}
	}
	float scale = (float)(max - min) / (float)255;
	buffer = HeightBuffer.pixels;
	for (i=0; i<512; i++) {
		for (int j=0; j<512; j++) {
			float scaledVal = ((float)(*buffer) - (float)min)/(float)scale;
			heightMap[i*512+j]=(unsigned char)scaledVal;
			buffer++;
		}
	}


}//@CODE_1192


void VoxelDemo::InitTables()
{//@CODE_1191

	for (int ang=0; ang<ANGLE_360; ang++) {
		float rad_angle = (float)ang*ANGULAR_INCREMENT*DEG_TO_RAD;
		sin_look[ang] = (float)sin(rad_angle);
		cos_look[ang] = (float)cos(rad_angle);
		if (ang < ANGLE_30) {
			sphere_cancel[ang+ANGLE_30] = 1/(float)cos(rad_angle);
			sphere_cancel[ANGLE_30-ang] = 1/(float)cos(rad_angle);
		}
	}
	for (int i=0; i<numberOfSlices; i++) {
		float delta = (float)play_z - (screen_z + (DISP_Y - nearScreenRow) + i); 
		distInfo[i].rayLength = ((float)(play_dist*play_z)/delta);
		distInfo[i].numRays = DISP_X; // until we gone use line thickness
	}
	// calculate stepsizes along the outer left and outer right rays
	// outer left & right are angled at viewAngle/2 from the player's view vector
	farStepSize = (distInfo[numberOfSlices-1].rayLength - distInfo[numberOfSlices-2].rayLength)/(float)cos(viewAngle/2.0f);
	nearStepSize = (distInfo[1].rayLength - distInfo[0].rayLength)/(float)cos(viewAngle/2.0f);
	float stepInc = (farStepSize - nearStepSize) / (float)(numberOfSlices - 1);
	float stepSize = nearStepSize;
	for (i=0; i<numberOfSlices; i++) {
		distInfo[i].stepSize = stepSize;
		stepSize += stepInc;
	}

	for (i=0; i<numberOfSlices; i++) {
		for (int j=0; j<256; j++) 
			distInfo[i].perspectiveLookup[j] = (int)mountain_scale * j / (int)(distInfo[i].rayLength+1);
	}
	


}//@CODE_1191


// End of Group: Initialization Methods

void VoxelDemo::DrawTerrain(int play_x, int play_y, int play_z, int play_ang,
                            int play_dist, Pixel* buffer, Pixel* texture,
                            unsigned char* height)
{//@CODE_1182

	int left_ang,right_ang,
		x_fine,y_fine,
		scale,
		top,
		diff,i;
	int x,y;  // world coordinates for height & texture
	float fx,fy;
	float ray_length;
	float ulx,uly,urx,ury; // world vectors ul and ur for edge stepping, these are unit vectors
	float wlx,wly,wrx,wry; // the cumulated vectors along the left/right edge
	float dx,dy;

	left_ang = play_ang - ANGLE_30;
	if (left_ang < 0) left_ang += ANGLE_360;
	right_ang = play_ang + ANGLE_30;
	if (right_ang > ANGLE_360) right_ang -= ANGLE_360;

	ulx = cos_look[left_ang];
	uly = sin_look[left_ang];
	urx = cos_look[right_ang];
	ury = sin_look[right_ang];

	ray_length = distInfo[0].rayLength;

	wlx = (float)play_x + ulx*ray_length;
	wly = (float)play_y + uly*ray_length;
	wrx = (float)play_x + urx*ray_length;
	wry = (float)play_y + ury*ray_length;

	for(i=0; i<DISP_X; i++) last_top[i] = 0xFFFF;  // optimize with memset

	for (int slice=0; slice<numberOfSlices; slice++) {
		TDistInfo info = distInfo[slice];

		dx = (wrx - wlx) / (float)info.numRays;
		dy = (wry - wly) / (float)info.numRays;
		fx = wlx;
		fy = wly; 

		for (int col=0; col < DISP_X; col++) {

			x = (int)fx;
			y = (int)fy;

			x_fine = x & 0x1FF;;
			y_fine = y & 0x1FF;

			Pixel tPix = texture[(x_fine << 9) + y_fine];
			unsigned char hPix = height[(x_fine << 9) + y_fine];

			//scale = (int)mountain_scale * hPix / (int)(info.rayLength+1);
			scale = info.perspectiveLookup[hPix];

			if (scale < 1) scale = 1;
			top = nearScreenRow-slice-scale;
			if (top < 0) top = 0;
			diff = last_top[col] - top;
			if (diff > 0) {
				if (top+diff > nearScreenRow) diff = (nearScreenRow - top);
				Pixel *offset;
				offset = buffer + (DISP_X * top) + col;
				for (i=diff; i; i--) {
					*offset = tPix;
					offset += DISP_X;
				}
			}
			last_top[col] = top;

			fx += dx;
			fy += dy;
		}
		wlx += info.stepSize*ulx;
		wly += info.stepSize*uly;
		wrx += info.stepSize*urx;
		wry += info.stepSize*ury;
	}



}//@CODE_1182


void VoxelDemo::GameEnd()
{//@CODE_1195

	TextureData->Unlock();
	HeightData->Unlock();
	free(TextureData);

}//@CODE_1195


bool VoxelDemo::GameInit()
{//@CODE_1193

	m_display = GetDisplay();
	HeightData = LoadImage(m_display, IDB_BITMAP3);
	HeightData->Lock(HeightBuffer);
	TextureData = LoadImage(m_display, IDB_BITMAP2);
	TextureData->Lock(TextureBuffer);
	InitTables();
	InitHeightMap();
	//hFile = ::CreateFile(_T("\\temp\\output.txt"), GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0, 0 );
    
    
	return true;

}//@CODE_1193


void VoxelDemo::GameLoop()
{//@CODE_1194


    //Surface::LockInfo buffer;
	m_display->GetBackBuffer()->Lock( BackBuffer );


    int width  = m_display->GetWidth();
    int height = m_display->GetHeight();

	memset(BackBuffer.pixels,0x10,width*height*2);

	play_ang += 5;
	if (play_ang > ANGLE_360) play_ang = 0;
	
	DrawTerrain(play_x,play_y,play_z,play_ang,play_dist,BackBuffer.pixels,TextureBuffer.pixels,heightMap);

 	m_display->GetBackBuffer()->Unlock();   
    m_display->Update();

}//@CODE_1194


//{{AFX DO NOT EDIT CODE BELOW THIS LINE !!!

/*@NOTE_1178
Method which must be called first in a constructor.
*/
void VoxelDemo::ConstructorInclude()
{
}


/*@NOTE_1179
Method which must be called first in a destructor.
*/
void VoxelDemo::DestructorInclude()
{
}


// Methods for the relation(s) of the class

//}}AFX DO NOT EDIT CODE ABOVE THIS LINE !!!

//@START_USER3
//////////////////////////////////////////////////////////////////////////////
//
// WinMain - Entry point
//
//////////////////////////////////////////////////////////////////////////////

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPTSTR, int )
{
    _Module.Init( 0, hInstance );
    
    VoxelDemo game;
    
	printf("HELLO HELLO");


    game.Run();
    
    return 0;
}
